function CSG_Cube()
{
	return new ScriptObject()
	{
		class = "CSGOperator";
		operator = "Cube";
		isPrimitive = 1;
	};
}

function CSG_Cylinder()
{
	return new ScriptObject()
	{
		class = "CSGOperator";
		operator = "Cylinder";
		isPrimitive = 1;
	};
}

function CSG_Difference(%obj0, %obj1)
{
	return new ScriptObject()
	{
		class = "CSGOperator";
		operator = "difference";
		arg0 = %obj0;
		arg1 = %obj1;
	};
}

function CSG_Intersect(%obj0, %obj1)
{
	return new ScriptObject()
	{
		class = "CSGOperator";
		operator = "intersect";
		arg0 = %obj0;
		arg1 = %obj1;
	};
}

function CSG_Rotate(%obj, %for, %up)
{
	%for = vectorNormalize(%for);
	%up = vectorNormalize(vectorSub(%up, vectorScale(%for, vectorDot(%for, %up))));
	return new ScriptObject()
	{
		class = "CSGOperator";
		operator = "rotation";
		arg0 = %obj;
		arg1 = %for;
		arg2 = %up;
		arg3 = vectorCross(%for, %up);
	};
}

function CSG_Scale(%obj, %scale)
{
	return new ScriptObject()
	{
		class = "CSGOperator";
		operator = "scale";
		arg0 = %obj;
		arg1 = %scale;
	};
}

function CSG_Sphere(%radius)
{
	%obj = new ScriptObject()
	{
		class = "CSGOperator";
		operator = "sphere";
		isPrimitive = 1;
	};
	if (%radius $= "")
		return %obj;
	return CSG_Scale(%obj, %radius SPC %radius SPC %radius);
}

function CSG_Torus(%rad0, %rad1)
{
	%obj = new ScriptObject()
	{
		class = "CSGOperator";
		operator = "torus";
		isPrimitive = 1;
	};
	if (%rad1 $= "")
	{
		%obj.arg0 = %rad0;
		return %obj;
	}
	%obj.arg0 = %rad1 / %rad0;
	return CSG_Scale(%obj, %rad0 SPC %rad0 SPC %rad0);
}

function CSG_Translate(%obj, %shift)
{
	return new ScriptObject()
	{
		class = "CSGOperator";
		operator = "translate";
		arg0 = %obj;
		arg1 = %shift;
	};
}

function CSG_Union(%obj0, %obj1)
{
	return new ScriptObject()
	{
		class = "CSGOperator";
		operator = "union";
		arg0 = %obj0;
		arg1 = %obj1;
	};
}

function CSGOperator::build(%csg, %delete)
{
	%csg.deleteOnComplete = %delete;
	CSGBuild_Init(%csg);
}

function CSGOperator::containsPoint(%csg, %point)
{
	switch$(%csg.operator)
	{
		case "cube":
			for (%i = 0; %i < 3; %i++)
				if (getWord(%point, %i) > 1 || getWord(%point, %i) < -1)
					return 0;
			return 1;
		case "cylinder":
			if (getWord(%point, 2) > 1 || getWord(%point, 2) < -1)
				return 0;
			return vectorLen(setWord(%point, 2, 0)) <= 1;
		case "difference":
			return %csg.arg0.containsPoint(%point) && !%csg.arg1.containsPoint(%point);
		case "intersect":
			return %csg.arg0.containsPoint(%point) && %csg.arg1.containsPoint(%point);
		case "rotation":
			%y = vectorDot(%point, %csg.arg1);
			%z = vectorDot(%point, %csg.arg2);
			%x = vectorDot(%point, %csg.arg3);
			return %csg.arg0.containsPoint(%x SPC %y SPC %z);
		case "scale":
			for (%i = 0; %i < 3; %i++)
				%point = setWord(%point, %i, getWord(%point, %i) / getWord(%csg.arg1, %i));
			return %csg.arg0.containsPoint(%point);
		case "sphere":
			return vectorLen(%point) <= 1;
		case "torus":
			%closest = vectorNormalize(setWord(%point, 2, 0));
			return vectorDist(%closest, %point) <= %csg.arg0;
		case "translate":
			return %csg.arg0.containsPoint(vectorSub(%point, %csg.arg1));
		case "union":
			return %csg.arg0.containsPoint(%point) || %csg.arg1.containsPoint(%point);
	}
}

function CSGOperator::deleteTree(%csg)
{
	if(%csg.isPrimitive)
		%csg.delete();
	else
	{
		switch$(%csg.operator)
		{
			case "difference":
				%csg.isPrimitive = 1;
				%csg.arg0.deleteTree();
				%csg.arg1.deleteTree();
				%csg.delete();
			case "intersect":
				%csg.isPrimitive = 1;
				%csg.arg0.deleteTree();
				%csg.arg1.deleteTree();
				%csg.delete();
			case "rotation":
				%csg.isPrimitive = 1;
				%csg.arg0.deleteTree();
				%csg.delete();
			case "scale":
				%csg.isPrimitive = 1;
				%csg.arg0.deleteTree();
				%csg.delete();
			case "union":
				%csg.isPrimitive = 1;
				%csg.arg0.deleteTree();
				%csg.arg1.deleteTree();
				%csg.delete();
		}
	}
}

function CSGOperator::getBounds(%csg)
{
	switch$(%csg.operator)
	{
		case "cube":
			return "-1 -1 -1 1 1 1";
		case "cylinder":
			return "-1 -1 -1 1 1 1";
		case "difference":
			return %csg.arg0.getBounds();
		case "intersect":
			%bounds = "0 0 0 0 0 0";
			%bounds0 = %csg.arg0.getBounds();
			%bounds1 = %csg.arg1.getBounds();
			for (%i = 0; %i < 3; %i++)
				%bounds = setWord(%bounds, %i, getMax(getWord(%bounds0, %i), getWord(%bounds1, %i)));
			for (%i = 3; %i < 6; %i++)
				%bounds = setWord(%bounds, %i, getMin(getWord(%bounds0, %i), getWord(%bounds1, %i)));
			return %bounds;
		case "rotation":
			if (%csg.cachedBounds !$= "")
				return %csg.cachedBounds;
			%bounds = %csg.arg0.getBounds();
			for (%i = 0; %i < 8; %i++)
			{
				%pos = getWord(%bounds, 0 + (%i & 1 ? 3 : 0))
					SPC getWord(%bounds, 1 + (%i & 2 ? 3 : 0))
					SPC getWord(%bounds, 2 + (%i & 4 ? 3 : 0));
				%pos = vectorDot(%pos, %csg.arg3) SPC vectorDot(%pos, %csg.arg1) SPC vectorDot(%pos, %csg.arg2);
				if (%i == 0)
				{
					%minX = %maxX = getWord(%pos, 0);
					%minY = %maxY = getWord(%pos, 1);
					%minZ = %maxZ = getWord(%pos, 2);
				}
				else
				{
					%x = getWord(%pos, 0);
					%y = getWord(%pos, 1);
					%z = getWord(%pos, 2);
					if (%x < %minX)
						%minX = %x;
					if (%x > %maxX)
						%maxX = %x;
					if (%y < %minY)
						%minY = %y;
					if (%y > %maxY)
						%maxY = %y;
					if (%z < %minZ)
						%minZ = %z;
					if (%z > %maxZ)
						%maxZ = %z;
				}
				echo(%x SPC %y SPC %z @ "   " @ %minX SPC %minY SPC %minZ SPC %maxX SPC %maxY SPC %maxZ);
			}
			%bounds = %minX SPC %minY SPC %minZ SPC %maxX SPC %maxY SPC %maxZ;
			%csg.cachedBounds = %bounds;
			return %bounds;
		case "scale":
			%bounds = %csg.arg0.getBounds();
			for (%i = 0; %i < 6; %i++)
				%bounds = setWord(%bounds, %i, getWord(%bounds, %i) * getWord(%csg.arg1, %i % 3));
			return %bounds;
		case "sphere":
			return "-1 -1 -1 1 1 1";
		case "torus":
			%sumRadius = 1 + %csg.arg0;
			%vector = %sumRadius SPC %sumRadius SPC %csg.arg0;
			return vectorScale(%vector, -1) SPC %vector;
		case "translate":
			%bounds = %csg.arg0.getBounds();
			return vectorAdd(getWords(%bounds, 0, 2), %csg.arg1) SPC vectorAdd(getWords(%bounds, 3, 5), %csg.arg1);
		case "union":
			%bounds = "0 0 0 0 0 0";
			%bounds0 = %csg.arg0.getBounds();
			%bounds1 = %csg.arg1.getBounds();
			for (%i = 0; %i < 3; %i++)
				%bounds = setWord(%bounds, %i, getMin(getWord(%bounds0, %i), getWord(%bounds1, %i)));
			for (%i = 3; %i < 6; %i++)
				%bounds = setWord(%bounds, %i, getMax(getWord(%bounds0, %i), getWord(%bounds1, %i)));
			return %bounds;
	}
}